/*
DROP DATABASE GeneticAlgorithms
*/

CREATE DATABASE GeneticAlgorithms
GO

USE GeneticAlgorithms
GO

CREATE TABLE ObjectStatistics (
	ObjectNumber TINYINT NOT NULL,
	Mass NUMERIC(4,2) NOT NULL,
	Volume NUMERIC(4,2) NOT NULL,
	Value NUMERIC(4,2) NOT NULL,
	NumberAvailable TINYINT NOT NULL,
	CONSTRAINT pk_ObjectStats PRIMARY KEY CLUSTERED (ObjectNumber)
);

CREATE TABLE BagStatistics (
	MaxMass NUMERIC(5,2),
	MaxVolume NUMERIC(5,2)
);

CREATE TABLE Individuals (
	Generation SMALLINT,
	IndividualID SMALLINT,
	CONSTRAINT pk_Individuals PRIMARY KEY CLUSTERED (Generation, IndividualID)
);
GO

CREATE TABLE IndividualDetails (
	Generation SMALLINT,
	IndividualID SMALLINT,
	ObjectNumber TINYINT,
	Quantity TINYINT,
	CONSTRAINT pk_IndividualDetails PRIMARY KEY CLUSTERED (Generation, IndividualID, ObjectNumber),
	CONSTRAINT fk_IndividualDetails_Individuals FOREIGN KEY (Generation, IndividualID) REFERENCES dbo.Individuals (Generation, IndividualID)
);
GO

-- used in the decoding of the binarystring.
CREATE TABLE Numbers (
	Value TINYINT PRIMARY KEY,
	BinaryRepresentation CHAR(8) UNIQUE
);
GO

CREATE TABLE GenerationStatistics (
	Generation SMALLINT,
	MaxFitness NUMERIC(8,2),
	AvgFitness NUMERIC(8,2)
);
GO

CREATE VIEW RandomNumber 
AS
SELECT RAND(CHECKSUM(NEWID())) AS Random;
GO

-- Modified from http://mitch-wheat.blogspot.com/2006/10/t-sql-way-converting-integers-to.html
CREATE FUNCTION IntegerToBinary(@intval int)
RETURNS char(8)
AS
BEGIN
	DECLARE @bincode char(64);
	SET @bincode = '0000000100100011010001010110011110001001101010111100110111101111';
	RETURN (
	SUBSTRING( @bincode, ((@IntVal / 16) & 15) * 4 + 1 , 4 ) +
	SUBSTRING( @bincode, (@IntVal & 15) * 4 + 1 , 4 )
	);
END
GO

CREATE FUNCTION dbo.ComputeBinaryString(@Generation SMALLINT, @Individual SMALLINT)
RETURNS CHAR(64)
AS
BEGIN
	DECLARE @BinaryString VARCHAR(64) = '';
	SELECT @BinaryString = @BinaryString + dbo.IntegerToBinary(Quantity) 
		FROM dbo.IndividualDetails
		WHERE Generation = @Generation AND IndividualID = @Individual
		ORDER BY ObjectNumber;
	RETURN @BinaryString;
END;
GO

CREATE FUNCTION dbo.Fitness(@Generation SMALLINT, @Individual SMALLINT)
RETURNS NUMERIC(6,2)
AS
BEGIN
	DECLARE @MaxMass NUMERIC(5,2), @MaxVolume NUMERIC(5,2);
	SELECT @MaxMass = MaxMass, @MaxVolume = MaxVolume
		FROM dbo.BagStatistics;
		
	DECLARE @TotalValue NUMERIC(6,2) = 0;
	
	SELECT @TotalValue = SUM(id.Quantity*os.Value)
		FROM dbo.IndividualDetails id
			INNER JOIN dbo.ObjectStatistics os ON id.ObjectNumber = os.ObjectNumber
		WHERE Generation = @Generation AND IndividualID = @Individual
		HAVING SUM(id.Quantity*os.Mass) <= @MaxMass 
			AND SUM(id.Quantity*os.Volume) <= @MaxVolume
			AND SUM(CASE WHEN id.Quantity>os.NumberAvailable THEN 1 ELSE 0 END) = 0;
			
	RETURN @TotalValue;
END;
GO

CREATE FUNCTION dbo.Crossover(@BinaryString1 CHAR(64), @BinaryString2 CHAR(64))
RETURNS @NewIndividuals TABLE (CrossoverPoint TINYINT, BinaryString CHAR(64))
AS
BEGIN

	-- where in the string is the crossover
	DECLARE @CrossoverPoint TINYINT;
	SELECT @CrossoverPoint = FLOOR(Random*64)+1 
		FROM RandomNumber;
	
	INSERT INTO @NewIndividuals
	SELECT @CrossoverPoint, SUBSTRING(@BinaryString1,0,@CrossoverPoint) + SUBSTRING(@BinaryString2,@CrossoverPoint, 64-@CrossoverPoint+1)
	UNION ALL
	SELECT @CrossoverPoint, SUBSTRING(@BinaryString2,0,@CrossoverPoint) + SUBSTRING(@BinaryString1,@CrossoverPoint, 64-@CrossoverPoint+1);
	
	RETURN
	
END
GO

CREATE FUNCTION dbo.ApplyMutation(@BinaryString CHAR(64))
	RETURNS CHAR(64)
AS
BEGIN
	DECLARE @MutationLocation TINYINT, @SwappedBit CHAR(1);
	
	-- Which bit are we toggling? 
	SELECT @MutationLocation = FLOOR(Random*64)+1 
		FROM RandomNumber;

	-- simple toggle for the chosen bit
	SET @SwappedBit = 1 - CAST(SUBSTRING(@BinaryString, @MutationLocation,1) AS BIT);
	
	-- and stuff it back in
	SET @BinaryString = STUFF(@BinaryString,@MutationLocation,1,@SwappedBit);
		
	RETURN @BinaryString;
END
GO

ALTER TABLE dbo.Individuals
	ADD BinaryString AS dbo.ComputeBinaryString (Generation, IndividualID),
		Fitness AS dbo.Fitness(Generation, IndividualID);
GO

CREATE PROCEDURE Evolve (@CurrentGeneration INT)
AS
BEGIN
	
	CREATE TABLE #SelectedIndividuals (
		BinaryString CHAR(64),
		CrossoverGroup TINYINT
	);
	
	-- use the fittest half of the current population to evolve the next generation
	-- ensures that the low fitness individuals are eliminated (survival of the fittest FTW)
	INSERT INTO #SelectedIndividuals (BinaryString, CrossoverGroup)
	SELECT BinaryString, NTILE(2) OVER (ORDER BY (SELECT NEWID())) AS CrossoverGroup
	FROM (
		SELECT TOP (50) PERCENT BinaryString
			FROM Individuals
			ORDER BY Fitness DESC
		) sub;
			
	-- Join so that I have the two binary strings I'm crossing over in the same row. 
	-- The function then does the cross over and the mutation applies to the result
	SELECT @CurrentGeneration AS Generation, ROW_NUMBER() OVER (ORDER BY (SELECT 1)) AS IndividualID, dbo.ApplyMutation(co.BinaryString) AS NewIndividual
	FROM
		(SELECT BinaryString, ROW_NUMBER() OVER (ORDER BY (SELECT NEWID())) AS PartnerNo
			FROM #SelectedIndividuals
			WHERE CrossoverGroup = 1) AS Group1
	INNER JOIN 
		(SELECT BinaryString, ROW_NUMBER() OVER (ORDER BY (SELECT NEWID())) AS PartnerNo
			FROM #SelectedIndividuals
			WHERE CrossoverGroup = 2) AS Group2
		ON Group1.PartnerNo = Group2.PartnerNo
	CROSS APPLY dbo.CrossOver(Group1.BinaryString, Group2.BinaryString) co;
	
END;
GO

-- Defined values

INSERT INTO dbo.Numbers (Value, BinaryRepresentation)
SELECT Value, dbo.IntegerToBinary(Value) FROM (
SELECT TOP (256) Row_Number() OVER (ORDER BY column_id)-1 AS Value FROM sys.columns) sub;
GO

INSERT INTO dbo.ObjectStatistics (ObjectNumber, Mass, Volume, Value, NumberAvailable)
VALUES  
	(1,0.5,0.45,2,50),
	(2,1.5,0.25,20,10),
	(3,0.1,1.25,8,150),
	(4,0.25,0.1,1,250),
	(5,0.67,0.3,6,100),
	(6,0.34,0.75,18,5),
	(7,1.25,0.25,5,40),
	(8,1.1,0.8,10,25);

INSERT INTO dbo.BagStatistics (MaxMass, MaxVolume)
VALUES  (100, 75);
GO

-- Main loop

SET NOCOUNT ON
DECLARE @CurrentGeneration SMALLINT = 0;
DECLARE @TotalGenerations SMALLINT = 50;
DECLARE @GenerationSize SMALLINT = 64; -- power of 2 best.

-- Create initial generation
-- First the individuals
INSERT INTO Individuals (Generation, IndividualID)
SELECT TOP(@GenerationSize) 
	@CurrentGeneration AS Generation, ROW_NUMBER() OVER (ORDER BY column_id) AS IndividualID
FROM sys.columns;

-- Now the details
INSERT INTO IndividualDetails (Generation, IndividualID, ObjectNumber, Quantity)
SELECT Generation, IndividualID, ObjectNumber, CEILING(RAND(CHECKSUM(NEWID()))*NumberAvailable/4)
	FROM Individuals CROSS JOIN ObjectStatistics;

-- Initial generation statistics
INSERT INTO GenerationStatistics (Generation, MaxFitness, AvgFitness)
SELECT @CurrentGeneration, MAX(Fitness), AVG(Fitness)
FROM Individuals;

-- placeholder table for new generations
CREATE TABLE #NewIndividuals (
	Generation SMALLINT,
	IndividualID SmallINT,
	BinaryString CHAR(64)
);

WHILE (@CurrentGeneration<@TotalGenerations)
	BEGIN

		SET @CurrentGeneration+=1; 
		
		-- Generate the new generation
		INSERT INTO #NewIndividuals (Generation, IndividualID, BinaryString)
		EXEC Evolve @CurrentGeneration;
		
		-- store the new individuals
		INSERT INTO Individuals (Generation, IndividualID)
		SELECT Generation, IndividualID 
		FROM #NewIndividuals;

		-- chop binary string, insert into details table.
		INSERT INTO IndividualDetails (Generation, IndividualID, ObjectNumber, Quantity)
		SELECT Generation, IndividualID, Allele, Value FROM
		(SELECT Generation, IndividualID, Value+1 AS Allele, SUBSTRING(BinaryString, 1+(Value*8), 8) AS BinaryString
			FROM #NewIndividuals
				CROSS JOIN Numbers 
			WHERE Value < 8) sub INNER JOIN Numbers ON sub.BinaryString = Numbers.BinaryRepresentation;

		-- delete all but the top (@GenerationSize) fittest individuals. Keep fixed population size. 
		DELETE FROM IndividualDetails 
			FROM IndividualDetails LEFT OUTER JOIN (
			SELECT TOP (@GenerationSize) Generation, IndividualID 
				FROM Individuals  
				ORDER BY fitness DESC
		) Chosen ON IndividualDetails.Generation = Chosen.Generation AND IndividualDetails.IndividualID = Chosen.IndividualID
		WHERE Chosen.IndividualID IS NULL;

		DELETE FROM Individuals WHERE BinaryString = ''; -- only possible when there are no details

		-- Compute current max fitness and current avg fitness
		INSERT INTO GenerationStatistics (Generation, MaxFitness, AvgFitness)
		SELECT @CurrentGeneration, MAX(Fitness), AVG(Fitness)
		FROM Individuals;
		
		-- cleanup
		TRUNCATE TABLE #NewIndividuals;
	END
DROP TABLE #NewIndividuals

GO

-- Results

SELECT TOP (1) i.Generation, i.IndividualID, TotalMass, TotalVolume, TotalValue 
	FROM dbo.Individuals AS i 
		INNER JOIN (
			SELECT Generation, IndividualID, SUM(Quantity*Value) AS TotalValue,  SUM(Quantity*Mass) AS TotalMass, SUM(Quantity*Volume) AS TotalVolume
				FROM dbo.IndividualDetails d INNER JOIN dbo.ObjectStatistics AS os ON d.ObjectNumber = os.ObjectNumber
				GROUP BY Generation, IndividualID
		) id ON i.Generation = id.Generation AND i.IndividualID = id.IndividualID
ORDER BY Fitness desc

SELECT  Generation ,
        MaxFitness ,
        AvgFitness FROM GenerationStatistics

